% Exemple de filtre a particule pour robot 2D avec un repere fixe
% et une distance.
% Nous allons utiliser plus de particules que necessaire au debut,
% pour representer le cas ou on ne connait rien de la position du robot.
% (c) 2010-2011 Philippe Giguere

clearvars
clf
myFontSize = 14;
dT = 0.2;  % Intervalle de temps entre les iterations du filtre

nStep = 40; % Nombre d'iterations

% Variables inconnues dcrivant le systme.
SBeacon  = 0.01; % Erreur en position pour la distance du repere
Somega   = 0.01; % Bruit veritable sur la vitesse angulaire, en rad/s
SV       = 0.02; % Bruit veritable sur la vitesse lineaire, en m/s
xVrai    = [-1 0 0]';  %

% Position des beacons (point de repere)
Xb       = [1 2 0];    
Yb       = [1 -1 0];   

% Commandes U envoyees au systeme 
V     = 0.4; % Vitesse lineaire, en m/s
omega = 0.0; % Vitesse angulaire, rad/s;

% Valeur des bruits utilises dans le filtre a particules
SVEstime = 10*SV;   % Valeur du bruit sur la vitesse lineaire pour deplacement des particules
SOmegaEstime = 1.0; % Valeur du bruit sur la vitesse angulaire pour deplacement des particules
SlaserEstime = 0.1; % Valeur du bruit sur laser estime dans le filtre particules

nParticules = 50;

% Je vais initialiser au hasard les particules, dans une boite de 5m par 5m
% centree sur (0,0)
X(1,:) = xVrai(1)*ones(1,nParticules); % position en x
X(2,:) = xVrai(2)*ones(1,nParticules); % position en y
X(3,:) = 2*pi*rand(1,nParticules)-pi;  % orientation
w = ones(1,nParticules)/nParticules; % Poids egaux pour les particules

C = [-2.5 2.5 -1.5 1.5]; % Les axes utilises dans les graphiques
clf
PlotParticules2D(X,w,'o',0.1);
axis(C); axis equal;
drawnow();

% ============== Boucle du filtre a particules =================
for iStep = 1:nStep

    % ======= Debut de la simulation du systeme reel =============
    time = iStep*dT;

    % Je simule la vraie position du robot
    xVrai(1) = xVrai(1) + (V+SV*randn)*cos(xVrai(3))*dT;
    xVrai(2) = xVrai(2) + (V+SV*randn)*sin(xVrai(3))*dT;
    xVrai(3) = xVrai(3) + (omega+Somega*randn)*dT;
    
    % Je lit la distace au repere
    for index=1:size(Xb,2)
        z(index)  = sqrt((xVrai(1)-Xb(index))^2 + (xVrai(2)-Yb(index))^2) + SBeacon*randn;
    end
    % ======= Fin de la simulation du systeme reel =============
    
    % ================ Debut du filtre a particules =============    
    for iParticule = 1:nParticules
        % Je propage chaque particule
        X(1,iParticule) = X(1,iParticule) + (V+SVEstime*randn)*cos(X(3,iParticule))*dT;
        X(2,iParticule) = X(2,iParticule) + (V+SVEstime*randn)*sin(X(3,iParticule))*dT;
        X(3,iParticule) = X(3,iParticule) + (omega+SOmegaEstime*randn)*dT;
        
        % Mise-a-jour des poids
        for index = 1:size(Xb,2)
            wnew = gauss(sqrt((X(1,iParticule)-Xb(index))^2+(X(2,iParticule)-Yb(index))^2)-z(index),SlaserEstime);
            w(iParticule) = w(iParticule).*wnew;
        end

    end
    [X w] = ParticuleResampling(X,w,0.5);
    % ================ Fin du filtre a particules =============

    % Montrer la position et le poids des particules
    clf
    PlotParticules2D(X,w,'o',0.1);
    hold on;
    
    %Position moyenne ponderee du robot
    Xmean = sum(X(1,:).*w);
    Ymean = sum(X(2,:).*w);
    Hmean = sum(X(3,:).*w);
    plot(Xmean,Ymean,'go','MarkerFaceColor','g','MarkerSize',7);
    l = line([Xmean  Xmean+0.2*cos(Hmean)],[Ymean Ymean+0.2*sin(Hmean)]);
    set(l,'Color','g','LineWidth',3);
    
    % Tracer les trois cercles represetant les contraintes
    for index = 1:size(Xb,2)
        h = circle([Xb(index) Yb(index)],z(index),1000);
        set(h,'LineWidth',2);
        plot(Xb(index),Yb(index),'rs','MarkerFaceColor','r');
    end
    
    % Position veritable du robot
    plot(xVrai(1),xVrai(2),'r+','MarkerSize',30);
    
    title(sprintf('Temps = %.2f s',time),'FontSize',myFontSize);
    set(gca,'FontSize',myFontSize-2);
    axis equal
    axis(C);
    pause(1); % Pour permettre de voir l'affichage
end